import os


def make(env_name, custom_args=None, path_to_rl=None):
    # Imports are inside the function because setup fails otherwise (Missing win32file)
    from rlgym.gym import Gym
    from rlgym.envs import match_factory

    match = match_factory.build_match(env_name, custom_args=custom_args)
    if match is None:
        raise ValueError("RLGym was unable to construct match!\n"
                     "Match ID: {}\n"
                     "Custom match params: {}".format(env_name, custom_args))

    return Gym(match, pipe_id=os.getpid(), path_to_rl=path_to_rl)


# FIXME i think this doesn't work [env = Gym]
def make_distributed(env_names, custom_arg_dicts=None, path_to_rl=None):
    from rlgym.gym import Gym
    from rlgym.distributed_gym import DistributedGym
    from rlgym.envs import match_factory

    envs = []
    for i in range(len(env_names)):
        name = env_names[i]
        args = custom_arg_dicts[i]
        match = match_factory.build_match(name, custom_args=args)
        if match is not None:
            env = Gym(match, pipe_id=i)
            envs.append(env)
        else:
            RuntimeError("RLGym was unable to construct match while building matches for distributed env!"
                  "\nMatch ID: {}"
                  "\nCustom match params: {}".format(name, args))

    if len(envs) == 0:
        raise RuntimeError("RLGym was unable to build any matches for distributed env!")

    return DistributedGym(envs)


from rlgym.utils import math

class PhysicsObject(object):
    def __init__(self, position=None, quaternion=None, linear_velocity=None, angular_velocity=None, orientation=None):
        self.position = position
        self.quaternion = quaternion
        self.linear_velocity = linear_velocity
        self.angular_velocity = angular_velocity
        self.orientation = orientation

    def decode_car_data(self, car_data):
        self.position = car_data[:3]
        self.quaternion = car_data[3:7]
        self.linear_velocity = car_data[7:10]
        self.angular_velocity = car_data[10:]
        #roll, pitch, yaw
        self.orientation = math.quat_to_orientation(self.quaternion)

    def decode_ball_data(self, ball_data):
        self.position = ball_data[:3]
        self.linear_velocity = ball_data[3:6]
        self.angular_velocity = ball_data[6:9]
        
    def serialize(self):
        repr = []

        if self.position is not None:
            for arg in self.position:
                repr.append(arg)
                
        if self.quaternion is not None:
            for arg in self.quaternion:
                repr.append(arg)
                
        if self.linear_velocity is not None:
            for arg in self.linear_velocity:
                repr.append(arg)
                
        if self.angular_velocity is not None:
            for arg in self.angular_velocity:
                repr.append(arg)

        if self.orientation is not None:
            for arg in self.orientation:
                repr.append(arg)

        return repr
        


from rlgym.envs import Match
from rlgym.utils import common_values
from rlgym.utils.terminal_conditions import common_conditions
from rlgym.utils.reward_functions import ShootBallReward
from rlgym.utils.obs_builders import RhobotObs


def basic_duels_match(self_play=False, custom_args=None):
    if custom_args is None:
        custom_args = {"ep_len_minutes": 5}

    keys = custom_args.keys()
    if "ep_len_minutes" not in keys:
        custom_args["ep_len_minutes"] = 5

    game_speed, tick_skip, spawn_opponents, random_resets, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(custom_args=custom_args)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)

def basic_doubles_match(self_play=False, custom_args=None):
    if custom_args is None:
        custom_args = {"ep_len_minutes": 5}

    keys = custom_args.keys()
    if "ep_len_minutes" not in keys:
        custom_args["ep_len_minutes"] = 5
    if "team_size" not in keys:
        custom_args["team_size"] = 2

    game_speed, tick_skip, spawn_opponents, random_resets, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(custom_args=custom_args)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)

def basic_standard_match(self_play=False, custom_args=None):
    if custom_args is None:
        custom_args = {"ep_len_minutes": 5}

    keys = custom_args.keys()
    if "ep_len_minutes" not in keys:
        custom_args["ep_len_minutes"] = 5
    if "team_size" not in keys:
        custom_args["team_size"] = 3

    game_speed, tick_skip, spawn_opponents, random_resets, team_size, terminal_conditions, reward_fn, obs_builder \
        = get_default_params(custom_args=custom_args)

    return Match(team_size=team_size,
                 tick_skip=tick_skip,
                 game_speed=game_speed,
                 spawn_opponents=spawn_opponents,
                 random_resets=random_resets,
                 self_play=self_play,
                 reward_function=reward_fn,
                 terminal_conditions=terminal_conditions,
                 obs_builder=obs_builder)


def get_default_params(custom_args=None):
    ep_len_minutes = 45 / 60
    if custom_args is not None:
        if "ep_len_minutes" in custom_args.keys():
            ep_len_minutes = custom_args["ep_len_minutes"]

    game_speed = 100
    if custom_args is not None:
        if "game_speed" in custom_args.keys():
            game_speed = custom_args["game_speed"]

    tick_skip = 8
    if custom_args is not None:
        if "tick_skip" in custom_args.keys():
            tick_skip = custom_args["tick_skip"]

    ticks_per_sec = 120
    ticks_per_min = ticks_per_sec * 60
    max_ticks = int(round(ep_len_minutes * ticks_per_min / tick_skip))

    spawn_opponents = True
    if custom_args is not None:
        if "spawn_opponents" in custom_args.keys():
            spawn_opponents = custom_args["spawn_opponents"]

    random_resets = False
    if custom_args is not None:
        if "random_resets" in custom_args.keys():
            random_resets = custom_args["random_resets"]

    team_size = 1
    if custom_args is not None:
        if "team_size" in custom_args.keys():
            team_size = custom_args["team_size"]

    terminal_conditions = [common_conditions.TimeoutCondition(max_ticks), common_conditions.GoalScoredCondition()]
    if custom_args is not None:
        if "terminal_conditions" in custom_args.keys():
            terminal_conditions = custom_args["terminal_conditions"]

    reward_fn = ShootBallReward()
    if custom_args is not None:
        if "reward_fn" in custom_args.keys():
            reward_fn = custom_args["reward_fn"]

    obs_builder = RhobotObs()
    if custom_args is not None:
        if "obs_builder" in custom_args.keys():
            obs_builder = custom_args["obs_builder"]

    return game_speed, tick_skip, spawn_opponents, random_resets, team_size, terminal_conditions, reward_fn, obs_builder



import numpy as np

def get_dist(x,y):
    return np.subtract(x,y)

def vector_projection(vec, dest_vec, mag_squared=None):
    """Compute the vector projection of vec on to dest_vec."""
    if mag_squared is None:
        norm = vecmag(dest_vec)
        if norm == 0:
            return dest_vec
        mag_squared = norm * norm

    if mag_squared == 0:
        return dest_vec

    dot = np.dot(vec, dest_vec)
    projection = np.multiply(np.divide(dot, mag_squared), dest_vec)
    return projection

def scalar_projection(vec, dest_vec):
    """Compute the scalar projection of vec on to dest_vec."""
    norm = vecmag(dest_vec)

    if norm == 0:
        return 0

    dot = np.dot(vec, dest_vec) / norm
    return dot

def squared_vecmag(vec):
    x = np.linalg.norm(vec)
    return x*x

def vecmag(vec):
    norm = np.linalg.norm(vec)
    return norm

def unitvec(vec):
    return np.divide(vec, vecmag(vec))

def quat_to_orientation(quat):
    w, x, y, z = quat
    sinr_cosp = 2 * (w * x + y * z)
    cosr_cosp = 1 - 2 * (x*x + y*y)
    sinp = 2 * (w * y - z * x)
    siny_cosp = 2 * (w * z + x * y)
    cosy_cosp = 1 - 2 * (y * y + z * z)

    roll = np.arctan2(sinr_cosp, cosr_cosp)
    if abs(sinp) > 1:
        pitch = np.pi/2
    else:
        pitch = np.arcsin(sinp)
    yaw = np.arctan2(siny_cosp, cosy_cosp)


    return roll, pitch, yaw


from rlgym.envs import match_configs as conf
from rlgym.envs import Match

def build_match(match_id, custom_args=None):
    m = match_id.strip().lower()
    self_play = "self" in m

    preconfigured_match = None
    if "duel" in m:
        preconfigured_match = conf.basic_duels_match
    elif "double" in m:
        preconfigured_match = conf.basic_doubles_match
    elif "standard" in m:
        preconfigured_match = conf.basic_standard_match

    if preconfigured_match is not None:
        match = preconfigured_match(self_play=self_play, custom_args=custom_args)
    else:
        game_speed, tick_skip, spawn_opponents, random_resets, team_size, terminal_conditions, reward_fn, obs_builder \
            = conf.get_default_params(custom_args=custom_args)

        match = Match(team_size=team_size,
                     tick_skip=tick_skip,
                     game_speed=game_speed,
                     spawn_opponents=spawn_opponents,
                     random_resets=random_resets,
                     self_play=self_play,
                     reward_function=reward_fn,
                     terminal_conditions=terminal_conditions,
                     obs_builder=obs_builder)

    return match



from rlgym.utils.obs_builders import ObsBuilder
from rlgym.utils import common_values, math
import numpy as np

class RhobotObs(ObsBuilder):
    def __init__(self):
        super().__init__()
        self.obs_size = 66

    def reset(self, optional_data=None):
        pass

    def build_obs(self, state, optional_data=None):
        raise NotImplementedError

    def build_obs_for_player(self, player, state, optional_data=None):
        prev_actions = optional_data
        if prev_actions is None:
            print("ATTEMPTED TO BUILD RHOBOT OBS WITH NO PREV ACTIONS ARGUMENT!")
            raise AssertionError

        players = state.players
        if player.team_num == common_values.ORANGE_TEAM:
            player_car = player.inverted_car_data
            ball = state.inv_ball
        else:
            player_car = player.car_data
            ball = state.ball

        ob = [arg for arg in prev_actions]
        ob.append(int(player.has_flip))
        ob.append(int(player.boost_amount))
        ob.append(int(player.on_ground))


        ob += player_car.position
        ob += player_car.orientation
        ob += [np.sin(arg) for arg in player_car.orientation]
        ob += [np.cos(arg) for arg in player_car.orientation]
        yaw = player_car.orientation[2]
        angle_between_bot_and_target = np.arctan2(ball.position[1] - player_car.position[1],
                                                  ball.position[0] - player_car.position[0])
        
        angle_front_to_target = angle_between_bot_and_target - yaw
        ob.append(angle_front_to_target)
        ob += player_car.linear_velocity
        ob += player_car.angular_velocity

        ob += ball.position
        ob += ball.linear_velocity
        ob += ball.angular_velocity

        ob += common_values.ORANGE_GOAL_CENTER
        ob += common_values.BLUE_GOAL_CENTER

        pb_dist = math.vecmag(math.get_dist(player_car.position, ball.position))
        ob.append(pb_dist)

        # FIXME this only works for the blue team
        pg_dist = math.vecmag(math.get_dist(player_car.position, common_values.ORANGE_GOAL_CENTER))
        ob.append(pg_dist)

        # FIXME this only works for the blue team
        pog_dist = math.vecmag(math.get_dist(player_car.position, common_values.BLUE_GOAL_CENTER))
        ob.append(pog_dist)

        for other in players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == common_values.BLUE_TEAM and player.team_num == other.team_num:
                car_data = other.car_data
            else:
                car_data = other.inverted_car_data

            # TODO: COMMENT THIS OUT
            #car_data = ObsBuilder.get_random_physics_state()

            ob += car_data.position
            ob += car_data.orientation
            ob += [np.sin(arg) for arg in car_data.orientation]
            ob += [np.cos(arg) for arg in car_data.orientation]
            ob += car_data.linear_velocity
            ob += car_data.angular_velocity

            pc_dist = math.vecmag(math.get_dist(player_car.position, car_data.position))
            ob.append(pc_dist)
            
            
        return ob